//  Robot_Driver.cc - a computer-controlled driver
//
//	Vamos Automotive Simulator
//  Copyright (C) 2008 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Robot_Driver.h"
#include "../body/Car.h"
#include "../body/Wheel.h"
#include "../geometry/Constants.h"
#include "../geometry/Numeric.h"
#include "../geometry/Three_Vector.h"
#include "../track/Strip_Track.h"
#include "World.h"

#include <limits>

using namespace Vamos_Body;
using namespace Vamos_Geometry;
using namespace Vamos_Track;
using namespace Vamos_World;

Robot_Driver::Robot_Driver (Car* car_in, Strip_Track* track_in)
 : Driver (car_in),
   m_steer_control (0.05, 0.0, 0.01),
   m_gas_control (50.0, 0.0, 0.0),
   m_slip_ratio (car_in->get_robot_parameters ().slip_ratio),
   m_deceleration (car_in->get_robot_parameters ().deceleration),
   m_lateral_acceleration (car_in->get_robot_parameters ().lateral_acceleration),
   m_road_index (0),
   m_segment_index (0),
   mp_track (track_in),
   m_shift_time (0.0),
   m_state (PARKED),
   m_state_time (0.0)
{
}

// Step the driver forward in time.
void 
Robot_Driver::propagate (double time_step) 
{
  switch (m_state)
    {
    case PARKED:
      car->brake (1.0);
      car->shift (0);
      car->disengage_clutch (0.0);
      if (car->engine ()->rotational_speed () < car->engine ()->stall_speed ())
        car->start_engine ();
      car->gas (0.0);

      m_state_time += time_step;
      if (m_state_time > 1.0)
        {
          car->brake (0.0);
          m_state_time = 0.0;
          m_state = STARTING;
        }
      return;

    case STARTING:
      if (m_state_time == 0.0)
        {
          car->engage_clutch (3.0);
          car->shift (1);
        }

      m_state_time += time_step;
      if (m_state_time > 3.0)
        {
          m_state_time = 0.0;
          m_state = DRIVING;
        }
      break;

    case DRIVING:
      break;
    }

  const Three_Vector track_position = 
    mp_track->track_coordinates (car->chassis ().position (), 
                                 m_road_index, 
                                 m_segment_index);

  steer (time_step, track_position);
  shift (time_step);
  accelerate (time_step, track_position);
}

void
Robot_Driver::draw ()
{
  // Uncomment to draw a point where the driver is aiming.
  // glLoadIdentity ();
  // glPointSize (8.0);
  // glColor3d (0.0, 0.8, 0.0);
  // glBegin (GL_POINTS);
  // Three_Vector target (target_position ());
  // glVertex3d (target.x, target.y, current_segment ().world_elevation (target));
  // glEnd ();
}

// The point that is kept on the racing line is this far ahead of the
// car.
double 
Robot_Driver::target_distance () const
{
  return std::max (15.0, 8.0 + 0.2 * speed ());
}

// The position of the point that the driver tries to keep on the racing line.
Three_Vector
Robot_Driver::target_position () const
{
  return car->chassis ().transform_to_world 
    (Three_Vector (target_distance (), car->center ().y, 0.0));
}

// Get the segment that the car is currently on.
const Gl_Road_Segment&
Robot_Driver::current_segment () const
{
  return *mp_track->get_road (m_road_index).segments ()[m_segment_index];
}

bool
Robot_Driver::on_curve (double along) const
{
  return ((current_segment ().radius () != 0.0) 
          && ((along - current_segment ().start_distance ()) 
              < 0.5 * current_segment ().length ()));
}

void
Robot_Driver::steer (double time_step, const Three_Vector& track_position)
{
  const Three_Vector ahead = 
    mp_track->track_coordinates (car->chassis ().position (), m_road_index, m_segment_index);

  const Three_Vector world_line 
    (mp_track->get_road (m_road_index).racing_line (ahead.x + target_distance ()));
  const Three_Vector goal = world_line - car->chassis ().position () + car->center ();
  const Three_Vector target = 
    target_position () - car->chassis ().position () + car->center ();

  car->steer (m_steer_control.output (target.cross (goal).z, time_step));
  //! take lateral slip into account.
}

void
Robot_Driver::shift (double time_step)
{
  if (m_state == STARTING)
    return;

  // Gear Selection
  int gear = car->transmission ()->gear ();
  double omega = car->engine ()->rotational_speed ();
  double up_omega = omega
    * car->transmission ()->gear_ratio (gear + 1)
    / car->transmission ()->gear_ratio (gear);
  double down_omega = omega
    * car->transmission ()->gear_ratio (gear - 1)
    / car->transmission ()->gear_ratio (gear);

  if (car->clutch ()->engaged ()
      && (omega < 1.1 * car->engine ()->stall_speed ()))
    {
      m_state = PARKED;
    }
  else if (m_shift_time > 1.0)
      m_shift_time = 0.0;
  else if (m_shift_time != 0.0)
      m_shift_time += time_step;
  else if ((gear < car->transmission ()->forward_gears ())
           && car->clutch ()->engaged ()
           && (car->engine ()->torque_map (1.0, omega)
               < car->engine ()->torque_map (1.0, up_omega)))
    {
      m_shift_time = time_step;
      car->shift_up ();
    }
  else if (car->clutch ()->engaged ()
           && (car->engine ()->torque_map (1.0, omega)
               < car->engine ()->torque_map (1.0, down_omega)))
    {
      m_shift_time = time_step;
      car->shift_down ();
    }
}

double
Robot_Driver::longitudinal_slip () const
{
  return std::max (car->wheel (0)->slip ().x,
                   std::max (car->wheel (1)->slip ().x,
                             std::max (car->wheel (2)->slip ().x, 
                                       car->wheel (3)->slip ().x)));
}

double
Robot_Driver::transverse_slip () const
{
  // Assume that the front wheels are steered.
  const double s1 = car->wheel (0)->slip ().y;
  const double s2 = car->wheel (1)->slip ().y;
  return (std::abs (s1) > std::abs (s2)) ? s1 : s2;
}

double
Robot_Driver::speed () const
{
  return car->chassis ().cm_velocity ().magnitude();
}

void
Robot_Driver::accelerate (double time_step, const Three_Vector& track_position)
{
  if (m_state == STARTING)
    {
      car->brake (0.0);    
      if (car->clutch ()->engaged ())
        {
          // Traction control
          car->gas (clip (m_gas_control.output (m_slip_ratio - longitudinal_slip (), 
                                                time_step), 
                          0.0, 1.0));
        }
      else
        {
          // Revs up
          car->gas (clip (m_gas_control.output (car->engine ()->peak_engine_speed ()
                                                - car->engine ()->rotational_speed (), 
                                                time_step), 0.0, 1.0));
        }
    }
  else
    change_speed (time_step, track_position);
}

// Return the minimum difference between the maximum speed on the
// racing line and the minimum possible future car speeds
double
Robot_Driver::velocity_margin ()
{
  double margin = std::numeric_limits <double>::max ();
  const double max_d = stopping_distance ();
  const double interval = 1.0;
  for (double d = 0.0; d < max_d; d += interval)
    {
      margin = std::min (margin, racing_line_speed(d) - stopping_speed (d));
    }
  return margin;
}

double 
Robot_Driver::deceleration () const
{
  return m_deceleration [0] + square (speed ()) * m_deceleration [1];
}

double 
Robot_Driver::lateral_acceleration () const
{
  return m_lateral_acceleration [0] + square (speed ()) * m_lateral_acceleration [1];
}

double
Robot_Driver::stopping_distance () const
{
  return 0.5 * square (speed ()) / deceleration ();
}

double
Robot_Driver::racing_line_speed (double distance)
{
  const Three_Vector position = car->chassis ().transform_to_world 
    (Three_Vector (car->center ().x, car->center ().y, 0.0));

  const Three_Vector ahead = 
    mp_track->track_coordinates (position, m_road_index, m_segment_index);

  const double curvature = mp_track->get_road (m_road_index).
    racing_line_curvature (ahead.x + distance);

  return std::sqrt (lateral_acceleration () / std::abs (curvature));
}

double
Robot_Driver::stopping_speed (double distance) const
{
  return std::sqrt (square (speed ()) - 2.0 * deceleration () * distance);
}

void
Robot_Driver::change_speed (double time_step, const Three_Vector& track_position)
{
  const double margin = 0.005 * velocity_margin ();
  if (margin < 0.0)
    {
      car->gas (0.0);
      car->brake (clip (-m_gas_control.output (margin, time_step), 0.0, 1.0));
    }
  else
    {
      car->gas (clip (m_gas_control.output ((margin < m_slip_ratio - longitudinal_slip ())
                                            ? margin
                                            : m_slip_ratio - longitudinal_slip (),
                                            time_step),
                      0.0, 1.0));
      car->brake (0.0);
    }
}
